// LU:
// Uses an LU decomposition to solve the linear set of equations, and
// to save memory, the L and U matrices are stored in an augmented A
// matrix

// Input:
// Problem rank n
// Parameter matrix a
// Vector b

// Output:
// Solution vector alpha

// Revised, Thomas Reppert 02/16/07

#include  <stdio.h>
#include <stdlib.h>
#include   <math.h>

void LU(int n, double** a, double* b, double* alpha)
{
	int 	i,j,k,m,ii,ip1,jm1,im1;
	double 	**A,sum,max,ab;
	
	A = (double **) calloc(n+1,sizeof(double *));
	for (i=0;i<=n;i++) {
		A[i] = (double *) calloc(n+2,sizeof(double));
		}
		
	/* make augmented a matrix */
	for (i=1;i<=n;i++) {
		A[i][n+1] = b[i];
		for (j=1;j<=n;j++)
			A[i][j] = a[i][j];
		}
		
	/* search first column of matrix for largest element */
	m=1;
	max=fabs(A[1][1]);
	for (i=2;i<=n;i++) {
		if (fabs(A[i][1]) > max) {
			max=fabs(A[i][1]);
			m=i;
			}
		}
		
	if (max == 0) {
		printf(" no unique solution exists \n");
		exit(1);
		}
	
	if (m != 1) {
		for (j=1;j<=n+1;j++) {
			max = A[m][j];
			A[m][j] = A[1][j];
			A[1][j] = max;
			}
		}
	/* calculate first row of the new matrix */
	for (j=2;j<=n+1;j++)
		A[1][j] = A[1][j]/A[1][1];
	/* calculate the ith  column from A[i][i] to A[n][i] */
	for (i=2;i<=n;i++) {
		j=i;
		for (ii=j;ii<=n;ii++) {
			sum=0.0;
			jm1=j-1;
			for (k=1;k<=jm1;k++)
				sum = sum + A[ii][k]*A[k][j];
			A[ii][j] = A[ii][j] - sum;
			}
		if (i != n) {
			/* search for largest A[ii][j] value from A[i][i] 
			   to A[n][i] */
			m=i;
			max=fabs(A[i][i]);
			ip1=i+1;
			for (ii=ip1;ii<=n;ii++) {
				ab = fabs(A[ii][i]);
				if (ab > max) {
					max = ab;
					m = ii;
					}
				}
			
			if (max == 0.0) {
				printf(" no solution m=%d\n",m);
				exit(1);
				}
			
			if ( m != i) {
				for (j=1;j<=n+1;j++) {
					max = A[m][j];
					A[m][j] = A[i][j];
					A[i][j] = max;
					}
				}
			}
		/* calculate the  new elements of the ith row from A[i][i+1]
		to A[i][n+1] */
		ip1 = i+1;
		for (j=ip1;j<=n+1;j++) {
			sum = 0.0;
			im1 = i-1;
			for (k=1;k<=im1;k++)
				sum = sum + A[i][k]*A[k][j];
			A[i][j] = (A[i][j]-sum)/A[i][i];
			}
		}
	/* solve for alpha by back substitution */
	alpha[n] = A[n][n+1];
	for (i=1;i<=n-1;i++) {
		sum = 0.0;
		for (j=n-i+1;j<=n;j++)
			sum = sum + A[n-i][j]*alpha[j];
		alpha[n-i] = A[n-i][n+1] - sum;
		}
		
	for (i=0;i<=n;i++) {
		free(A[i]);
		}
	free(A);
}
